import json
from datetime import datetime
from multiprocessing import cpu_count
from multiprocessing.pool import ThreadPool
from os import makedirs
from os.path import join
from typing import Dict, Optional

from pandas import DataFrame, Timestamp

from .factor_risk_model import FactorRiskModel
from .risk_model import RiskModel
from .rolling_risk_model import RollingRiskModel


class RollingFactorRiskModel(RollingRiskModel):
    """
    Rolling factor risk model.

    Rolling factor risk model is a subclass of rolling risk model,
    while it supports transforming each risk model given the instrument
    returns.
    """

    def __init__(
        self,
        model: Optional[RiskModel] = None,
        window: Optional[int] = None,
        show_progress: Optional[bool] = False,
        values: Optional[Dict[datetime, RiskModel]] = None,
    ):
        """
        Constructor.

        Parameters
        ----------
        model: Optional[RiskModel]
            Risk model object to fit in rolling basis.

        window: Optional[int]
            Number of rolling windows to use from the returns.
            Must be provided in fitting the model.

        show_progress: Optional[bool]
            Indicate to show progress bar in running.

        values: Optional[Dict[datetime, RiskModel]]
            Rolling risk models values.
        """
        super().__init__(
            model=model,
            window=window,
            show_progress=show_progress,
            values=values,
        )

    def transform(
        self,
        y: DataFrame,
        validity: Optional[DataFrame] = None,
        regressor: Optional[object] = None,
        start_date: Optional[Timestamp] = None,
    ) -> object:
        """
        Transform the rolling factor risk model.

        The method is used to transform the rolling factor risk model by
        passing another set of returns. Most of the time, the
        factor risk model is fitted by the estimation universe,
        and then transformed by the model universe.

        Parameters
        ----------
        y : DataFrame
            The instrument returns of which its index and columns
            are the date / time and return values.

        validity: DataFrame
            The instrument validity on the date.

        regressor : object, default=None
            Regressor to transform the input y into factor exposures.
            If None, the regressor is set to the default WLS.

        Returns
        -------
        object
            The transformed rolling factor risk model.
        """
        if not isinstance(y, DataFrame):
            raise TypeError(
                "Only DataFrame type is supported, but not " f"{y.__class__.__name__}"
            )

        if not self._config.window:
            raise ValueError(
                f"Rolling timeframe must be specified, but not {self._config.window}"
            )

        values = {}
        iterator = self.keys()
        if self._config.show_progress:
            from tqdm import tqdm

            iterator = tqdm(iterator, leave=False)

        for index in iterator:
            if start_date is not None and start_date > index:
                continue
            y_end_index = y.index.get_loc(index)
            y_start_index = y_end_index - self._config.window
            if y_start_index < 0:
                raise ValueError(
                    "Input data does not have sufficient history for " f"index {index}"
                )

            y_input = y.iloc[y_start_index : y_end_index + 1]
            if validity is not None:
                validity_input = validity.loc[index]
                y_input = y_input.loc[:, validity_input]

            # Skip if the number of sample size is zero
            if y_input.shape[1] == 0:
                continue

            risk_model = self.get(index)
            values[index] = risk_model.transform(
                y=y_input.fillna(0.0),
                regressor=regressor,
            )

        self._values = values
        return self

    def write_directory(
        self, path: str, format: str = "parquet", workers: int = cpu_count(), **kwargs
    ):
        """
        Write to a directory.

        Parameters
        ----------
        path: str
            Directory path to write to.

        format: str
            Supported formats. Default is "parquet". Options
            are "csv", "parquet" and "hdf".

        workers: int
            Number of workers to use for parallel write operations.
            Default is the number of CPUs provided.

        **kwargs: dict
            Optional keyword arguments for the write operation.
        """

        def _frm_write_directory(item):
            key, frm = item
            if isinstance(key, Timestamp):
                key = key.isoformat()
            key_path = join(path, key)
            makedirs(key_path, exist_ok=True)
            frm.write_directory(key_path, format=format, **kwargs)
            return key

        with ThreadPool(processes=workers) as pool:
            keys = pool.map(_frm_write_directory, self._values.items())

        with open(join(path, "metadata.json"), mode="w+") as fp:
            json.dump({"directories": keys, "parameters": self.asdict()}, fp)

    @classmethod
    def read_directory(
        cls, path: str, format: str = "parquet", workers: int = cpu_count(), **kwargs
    ):
        """
        Read a model from directory.

        Parameters
        ----------
        path: str
            Directory path to read from.

        format: str
            Supported formats. Default is "parquet". Options
            are "csv", "parquet" and "hdf".

        workers: int
            Number of workers to use for parallel read operations.
            Default is the number of CPUs provided.

        **kwargs: dict
            Optional keyword arguments for the read operation.
        """

        def _frm_read_directory(key):
            key_path = join(path, key)
            value = FactorRiskModel.read_directory(key_path, format=format, **kwargs)
            return Timestamp(key), value

        with open(join(path, "metadata.json")) as fp:
            metadata = json.load(fp)
            directories = metadata["directories"]
            metadata = metadata["parameters"]

        with ThreadPool(processes=workers) as pool:
            values = pool.map(_frm_read_directory, directories)

        return cls(values=dict(values), **metadata)



from abc import ABC, abstractmethod
from typing import Union

from numpy import diagonal, ndarray, sqrt
from pandas import DataFrame, Series

from .config import Config
from .engine import NumpyEngine

np = NumpyEngine()


class RiskModelConfig(Config):
    """
    Risk model configuration.

    Parameters
    ----------
    show_all_instruments : bool.
        Indicate whether to show all instruments. Default is False.
        If True, the instruments outside of the universe in each
        period may not be filtered out.
    """

    show_all_instruments: bool = False


class RiskModel(ABC):
    """
    Risk Model.

    The class is an abstract class to compute covariance and other
    related metrics.
    """

    ConfigClass = RiskModelConfig

    def __init__(self, show_all_instruments: bool = False, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        show_all_instruments : bool.
            Indicate whether to show all instruments. Default is False.
            If True, the instruments outside of the universe in each
            period may not be filtered out.
        """
        self._config = self.ConfigClass(
            show_all_instruments=show_all_instruments, **kwargs
        )

    @property
    def config(self):
        """
        Return the configuration object.
        """
        return self._config

    @abstractmethod
    def cov(self, **kwargs) -> ndarray:
        """
        Get the covariance matrix.

        Returns
        -------
        numpy.ndarray
            A square pairwise covariance matrix which its
            diagonal entries are the variances.
        """

    def vol(self, **kwargs) -> ndarray:
        """
        Get the volatility series.

        Returns
        -------
        numpy.ndarray
            Volatility series derived from covariance matrix.
        """
        cov = self.cov(**kwargs)
        vol = sqrt(diagonal(cov))
        if isinstance(cov, DataFrame):
            vol = Series(vol, index=cov.index)
        return vol

    def corr(self, **kwargs) -> ndarray:
        """
        Get the correlation matrix.

        Returns
        -------
        numpy.ndarray
            A square pairwise correlation matrix which its
            diagonal entries are all ones.
        """
        cov = self.cov(**kwargs)
        vol = np.sqrt(np.diagonal(cov))
        return ((cov / vol).T / vol).T

    def asdict(self):
        """
        Returns a dict representation of the object.
        """
        return self.config.dict()

    @staticmethod
    def _to_numpy(values: Union[ndarray, DataFrame]) -> ndarray:
        """
        Convert the values to a numpy array
        """
        if values is None:
            return values
        elif isinstance(values, (DataFrame, Series)):
            return np.array(values.values)
        elif isinstance(values, ndarray):
            return np.array(values)
        else:
            raise TypeError(
                "Expect either pandas DataFrame or numpy array, "
                f"but got {values.__class__.__name__}"
            )



from typing import Optional

from numpy import diag_indices_from, trace
from pandas import DataFrame, Series

from .risk_model import RiskModel
from .rolling_factor_risk_model import RollingFactorRiskModel


class CovarianceEstimator:
    """
    Covariance estimator.
    """

    def __init__(
        self,
        risk_model: RiskModel,
        shrinkage_method: Optional[str] = None,
        delta: Optional[float] = None,
    ):
        """
        Constructor.

        Parameters
        ----------
        risk_model : RiskModel
          Risk model object.
        shrinkage_method : Optional[str]
          Shrinkage method. Options are "constant" and "ledoit_wolf_constant_variance".
        delta : Optional[float]
          Delta, only used in constant shrinkage.
        """
        self._risk_model = risk_model
        self.shrinkage_method = shrinkage_method
        self.delta = delta

    def corr(self) -> DataFrame:
        """
        Correlation
        """
        return self._risk_model.corr()

    def cov(
        self, volatility: Optional[Series] = None, strict: bool = True
    ) -> DataFrame:
        """
        Correlation

        Parameters
        ----------
        volatility : pd.Series
          Volaility series to convert from correlation to covariance. Optional.
        strict : bool
          Indicates to throw exception if volatility series does not align with
          correlation matrix.
        """
        if volatility is None:
            return self._risk_model.cov()

        corr = self._risk_model.corr()
        if strict and set(volatility.index) != set(corr.index):
            raise ValueError(
                "Incorrect volatility series passed. Length of volatility "
                f"is {len(volatility.index)} while that of correlation is "
                f"{len(corr.index)}"
            )
        elif len(volatility.index) < len(corr.index):
            instruments = volatility.index
        else:
            instruments = corr.index

        volatility = volatility.loc[instruments]
        cov = (
            corr.loc[instruments, instruments]
            .mul(volatility, axis=0)
            .mul(volatility, axis=1)
        )

        if self.shrinkage_method == "constant":
            cov = self.constant_shrinkage(cov, self.delta)
        elif self.shrinkage_method is not None:
            raise ValueError(
                f"Cannot recognize shrinkage method {self.shrinkage_method}"
            )

        return cov

    @staticmethod
    def constant_shrinkage(cov: DataFrame, delta: float):
        """
        Constant shrinkage.
        """
        N = len(cov)
        avg_var = trace(cov) / N * delta
        cov *= 1 - delta
        cov.values[diag_indices_from(cov)] += avg_var
        return cov


class RollingCovarianceEstimator:
    """
    Rolling covariance estimator.
    """

    def __init__(
        self,
        rolling_risk_model: RollingFactorRiskModel,
        shrinkage_method: Optional[str] = None,
        delta: Optional[float] = None,
    ):
        """
        Constructor.

        Parameters
        ----------
        rolling_risk_model : RollingFactorRiskModel
          Rolling risk model object.
        shrinkage_method : Optional[str]
          Shrinkage method. Options are "constant" and "ledoit_wolf_constant_variance".
        delta : Optional[float]
          Delta, only used in constant shrinkage.
        """
        self._rolling_risk_model = rolling_risk_model
        self.shrinkage_method = shrinkage_method
        self.delta = delta

    def cov(self, volatility: Optional[DataFrame] = None):
        """
        Correlation

        Parameters
        ----------
        volatility : DataFrame
          Volaility series to convert from correlation to covariance. Optional.
        """
        return {
            date: (
                CovarianceEstimator(
                    risk_model,
                    shrinkage_method=self.shrinkage_method,
                    delta=self.delta,
                ).cov(volatility.loc[date], strict=False)
            )
            for date, risk_model in self._rolling_risk_model.items()
        }



from contextlib import contextmanager
from os import environ
from typing import Any

_BACKEND_ENGINE = "numpy"
_SUPPORTED_ENGINES = ["numpy", "tensorflow", "cupy", "jax", "torch", "dask"]


def backend():
    return _BACKEND_ENGINE


def set_backend(library_name):
    """
    Set backend engine.

    The function sets the backend engine in global level.

    Parameters
    ----------
    library_name : str
        Library name. Default is `numpy`. Options are `numpy`, `tensorflow`,
        `cupy`, `jax` and `torch`.
    """
    library_name = library_name.lower()
    if library_name not in _SUPPORTED_ENGINES:
        raise ValueError(
            "Only `numpy`, `tensorflow`, `cupy`, `jax`, `torch` and `dask` "
            f"are supported, but not {library_name}"
        )
    global _BACKEND_ENGINE
    _BACKEND_ENGINE = library_name
    return _BACKEND_ENGINE


@contextmanager
def use_backend(library_name="numpy"):
    """
    NumPy engine selection.

    The function is a context manager to enable users to switch to a
    specific library as a replacement of NumPy in CPU.

    Parameters
    ----------
    library_name : str
        Library name. Default is `numpy`. Options are `numpy`, `tensorflow`,
        `cupy`, `jax` and `torch`.
    """
    library_name = library_name.lower()
    if library_name not in _SUPPORTED_ENGINES:
        raise ValueError(
            "Only `numpy`, `tensorflow`, `cupy`, `jax`, `torch` and `dask` "
            f"are supported, but not {library_name}"
        )
    global _BACKEND_ENGINE
    _original = _BACKEND_ENGINE
    try:
        _BACKEND_ENGINE = library_name
        yield
    finally:
        _BACKEND_ENGINE = _original


class NumpyEngine:
    """
    NumPy engine.
    """

    def __getattribute__(self, __name: str) -> Any:
        if __name == "name":
            return _BACKEND_ENGINE

        try:
            if _BACKEND_ENGINE == "numpy":
                import numpy as anp
            elif _BACKEND_ENGINE == "tensorflow":
                import tensorflow.experimental.numpy as anp

                anp.experimental_enable_numpy_behavior()
            elif _BACKEND_ENGINE == "cupy":
                import cupy as anp
            elif _BACKEND_ENGINE == "jax":
                import jax.numpy as anp
            elif _BACKEND_ENGINE == "torch":
                import torch as anp

                anp.array = anp.tensor
                anp.ndarray = anp.Tensor
                anp.newaxis = None
            elif _BACKEND_ENGINE == "dask":
                import dask.array as anp
                from numpy import newaxis

                anp.newaxis = newaxis
            else:
                raise ValueError(f"Cannot recognize backend {_BACKEND_ENGINE}")
        except ImportError:
            raise ImportError(
                "Library `numpy` cannot be imported from backend engine "
                f"{_BACKEND_ENGINE}. Please make sure to install the library "
                f"via `pip install {_BACKEND_ENGINE}`."
            )

        try:
            return getattr(anp, __name)
        except AttributeError:
            raise AttributeError(
                f"Cannot get attribute / function ({__name}) from numpy library in "
                f"backend engine {_BACKEND_ENGINE}"
            )


class LinAlgEngine:
    """
    Linear algebra engine.
    """

    def __getattribute__(self, __name: str) -> Any:
        if __name == "name":
            return _BACKEND_ENGINE

        try:
            if _BACKEND_ENGINE == "numpy":
                import numpy.linalg as alinalg
            elif _BACKEND_ENGINE == "tensorflow":
                import tensorflow.experimental.numpy as anp
                import tensorflow.linalg as alinalg

                anp.experimental_enable_numpy_behavior()
            elif _BACKEND_ENGINE == "cupy":
                import cupy.linalg as alinalg
            elif _BACKEND_ENGINE == "jax":
                import jax.numpy.linalg as alinalg
            elif _BACKEND_ENGINE == "torch":
                import torch.linalg as alinalg
            elif _BACKEND_ENGINE == "dask":
                import dask.array.linalg as alinalg

                # XXX: fall back pinv to inv
                alinalg.pinv = alinalg.inv
            else:
                raise ValueError(f"Cannot recognize backend {_BACKEND_ENGINE}")
        except ImportError:
            raise ImportError(
                "Library `linalg` cannot be imported from backend engine "
                f"{_BACKEND_ENGINE}. Please make sure to install the library "
                f"via `pip install {_BACKEND_ENGINE}`."
            )

        try:
            return getattr(alinalg, __name)
        except AttributeError:
            raise AttributeError(
                f"Cannot get attribute / function ({__name}) from linalg library in "
                f"backend engine {_BACKEND_ENGINE}"
            )


set_backend(environ.get("FPM_BACKEND_ENGINE", "numpy"))
numpy = NumpyEngine()
linalg = LinAlgEngine()



from dataclasses import dataclass
from typing import Optional

from numpy import ndarray

from ..engine import LinAlgEngine

linalg = LinAlgEngine()


@dataclass
class RegressionResult:
    """
    Regression result.

    Assume that

      y = X @ beta + alpha

    where

      X is a matrix in shape (m, n)
      y is a matrix in shape (m, k)

    Then

      beta is a matrix in shape (n, k)
      alpha is a matrix in shape (m, k)

    Parameters
    ----------
    alpha: Optional[ndarray]
      The error term of the regression.
    beta: Optional[ndarray]
      The fitted linear variables of the regression.
    """

    alpha: Optional[ndarray] = None
    beta: Optional[ndarray] = None


class WLS:
    """
    Weighted least squares problem solver.

    The solver is to run regression with weighted least squares
    objective.
    """

    def __init__(self, executor: str = "closed"):
        """
        Construct

        Parameters
        ----------
        executor : str
          Executor name. Default is "closed", deriving from closed formula.
        """
        self._executor = executor

    def fit(self, X: ndarray, y: ndarray, weights: Optional[ndarray] = None):
        """
        Fit the coefficients by the data X and y.

        Parameters
        ----------
        X: ndarray
          Training data in dimension (m, n).
        y: ndarray
          Target values in dimension (m, k).
        weights: Optional[ndarray]
          Weightings in regressiond data. The dimension should be
          same as y.
        """
        if self._executor == "closed":
            return self._close_fit(X=X, y=y, weights=weights)

        raise ValueError(f"Executor {self._executor} is not supported")

    @staticmethod
    def _close_fit(X: ndarray, y: ndarray, weights: Optional[ndarray] = None):
        """
        Fit the coefficients with closed formula.

        coefficients = (X^T @ W @ X)^{-1} @ X^T @ W @ y
        """
        if isinstance(weights, ndarray):
            if len(weights.shape) == 1 and weights.shape[0] == y.shape[0]:
                weights = weights**0.5
                X_t_w = X.T * weights * weights.T
                beta = linalg.pinv(X_t_w @ X) @ X_t_w @ y
            else:
                raise ValueError(
                    f"Dimension of y {y.shape} does not align with weights "
                    f"{weights.shape}"
                )
        else:
            beta = linalg.pinv(X.T @ X) @ X.T @ y

        alpha = y - X @ beta
        return RegressionResult(alpha=alpha, beta=beta)



from typing import Optional, Union

from pandas import DataFrame, Series


def compute_forecast_returns(
    X: Union[DataFrame, Series],
    forecast_timeframe: int,
    min_periods: Optional[int] = None,
) -> Union[DataFrame, Series]:
    """
    Returns the forecast returns for a given forecast timeframe.

    Parameters
    ----------
    X : Union[DataFrame, Series]
        Dataframe of instrument returns.
    forecast_timeframe : int
        The forecast timeframe.
    min_periods : Optional[int]
        Minimum number of observations in the forecast.

    Returns
    ----------
    Union[DataFrame, Series]
        Sum of forecast returns.
    """
    return (
        X.rolling(forecast_timeframe, min_periods=min_periods)
        .mean()
        .shift(-forecast_timeframe)
    )



from pandas import DataFrame


def equal_weighted_portfolio(validity: DataFrame) -> DataFrame:
    """
    Construct an equal weighted portfolio from validity.

    Parameters
    ----------
    validity : DataFrame
        DataFrame containing the universe validity.

    Returns
    -------
    DataFrame
        DataFrame containing the weights of the equal weighted portfolio.
    """
    weights = DataFrame(
        1.0,
        index=validity.index,
        columns=validity.columns,
    )
    weights = weights.where(validity)
    weights = weights.div(weights.sum(axis=1), axis=0).fillna(0.0)
    return weights



import json
from os.path import join
from typing import Optional

from numpy import any, diag_indices_from, nan, ndarray
from pandas import DataFrame, Series

from .engine import NumpyEngine
from .regressor import WLS
from .risk_model import RiskModel

np = NumpyEngine()


class FactorRiskModel(RiskModel):
    """
    Factor Risk Model.

    The class is an abstract class to fit the factor risk model.

    The factor risk model contains the data attribute `factor_exposures`,
    `factors` and `residual_returns`.

    The factor exposures are the exposures of each instrument to the
    specified factors.

    The factor returns are returns among the date / time series for each
    factor.

    The residual returns are the idiosyncratic returns of the instruments
    regarding the specified factor exposures and returns.
    """

    def __init__(
        self,
        factor_exposures: ndarray = None,
        factor_returns: ndarray = None,
        residual_returns: ndarray = None,
        **kwargs,
    ):
        """
        Constructor

        Parameters
        ----------
        factor_exposures : ndarray
          Factor exposures of the factor risk model.
        factors_returns : ndarray
          Factor returns of the factor risk model.
        residual_returns : ndarray
          Residual returns of the factor risk model.
        """
        super().__init__(**kwargs)
        self._factor_exposures = factor_exposures
        self._factor_returns = factor_returns
        self._residual_returns = residual_returns

    @property
    def factor_exposures(self) -> ndarray:
        """
        Return the factor exposures.

        Return
        ------
        ndarray
          Matrix in dimension (n, N) where N is the number of
          instruments and n is the number of factors.
        """
        return self._factor_exposures

    @property
    def factor_returns(self) -> ndarray:
        """
        Return the factor returns.

        Return
        ------
        ndarray
          Matrix in dimension (T, n) where n is the number of
          factors and T is the number of time frames.
        """
        return self._factor_returns

    @property
    def residual_returns(self) -> ndarray:
        """
        Return the residual returns.

        Return
        ------
        ndarray
          Matrix in dimension (T, N) where N is the number of
          instruments and T is the number of time frames.
        """
        return self._residual_returns

    def fit(self, X: ndarray, weights: Optional[ndarray] = None) -> object:
        """
        Fit the model.

        Parameters
        ----------
        X : ndarray
          Input array of shape (T, N) where N is the number of
          instruments and T is the number of timeframes.

        weights: Optional[ndarray]
          Weights array of shape (N,) where N is the number of
          instruments.
        """
        pass

    def copy(self) -> object:
        """
        Copy the model.

        Returns
        -------
        object
          Copy of the model.
        """
        return FactorRiskModel(
            factor_exposures=self._factor_exposures.copy(),
            factor_returns=self._factor_returns.copy(),
            residual_returns=self._residual_returns.copy(),
            **self._config.dict(),
        )

    def specific_variances(self, weights=None, ddof=1) -> ndarray:
        """
        Get specific variances.

        Parameters
        ----------
        ddof : int
          Degrees of freedom.

        Returns
        -------
        ndarray
          Specific variances of the instruments.
        """
        T = self._residual_returns.shape[0]
        residual_returns = self._residual_returns
        if isinstance(self._residual_returns, DataFrame):
            residual_returns = residual_returns.values

        if weights is not None:
            r_mean = np.mean(residual_returns * weights[:, np.newaxis], axis=0)
            variances = (residual_returns - r_mean) ** 2
            variances *= weights[:, np.newaxis]
        else:
            r_mean = np.mean(residual_returns, axis=0)
            variances = (residual_returns - r_mean) ** 2

        variances = sum(variances) / (T - ddof)

        if isinstance(self._residual_returns, DataFrame):
            variances = Series(variances, index=self._residual_returns.columns)

        return variances

    def transform(self, y: ndarray, regressor: Optional[object] = None) -> object:
        """
        Transform the factor risk model.

        The method is used to transform the factor risk model by
        passing another set of returns. Most of the time, the
        factor risk model is fitted by the estimation universe,
        and then transformed by the model universe.

        Parameters
        ----------
        y : ndarray
            The instrument returns.

        regressor : object, default=None
            Regressor to transform the input y into factor exposures.
            If None, the regressor is set to the default WLS.

        Returns
        -------
        ndarray
            The transformed factor risk model.
        """
        X = self.factor_returns
        if X is None:
            raise ValueError("Factor returns must be initialised first")

        # Convert the factor returns into a ndarray first
        if isinstance(X, DataFrame):
            X = X.values

        # Convert the y input to a ndarray first
        y_input = self._to_numpy(y)

        # Set the default regressor
        regressor = regressor or WLS()

        # Transform the factor exposures from the y input
        regressor_result = regressor.fit(X=X, y=y_input)
        factor_exposures = regressor_result.beta
        residual_returns = regressor_result.alpha

        if isinstance(self.factor_returns, DataFrame):
            factor_exposures = DataFrame(
                factor_exposures,
                index=self.factor_exposures.index,
                columns=y.columns,
            )
            residual_returns = DataFrame(
                residual_returns,
                index=y.index,
                columns=y.columns,
            )

        self._factor_exposures = factor_exposures
        self._residual_returns = residual_returns
        return self

    def cov(self, halflife: Optional[float] = None, ddof=1) -> ndarray:
        """
        Get the covariance matrix.

        Parameters
        ----------
        halflife : Optional[float]
            Half life in applying the exponential weighting on factor
            returns for computing the factor covariance matrix. If
            None is passed, no exponential weighting is applied.

        Returns
        -------
        numpy.ndarray
            A square pairwise covariance matrix which its
            diagonal entries are the variances.
        """
        B = self._factor_exposures
        F = self._factor_returns
        if F is None:
            raise ValueError("Factor return cannot be None")
        elif isinstance(F, DataFrame):
            F = F.values

        W = None
        T = F.shape[0]
        if halflife is not None:
            W = np.array([2 ** (-(T - 1 - t) / halflife) for t in range(0, T)])
            F = F * (W[:, np.newaxis] ** 0.5)

        F = F - np.mean(F, axis=0)
        factor_covariances = (F.T @ F) / (T - ddof)
        specific_variances = self.specific_variances(weights=W, ddof=ddof)

        R = specific_variances
        if isinstance(B, DataFrame):
            instruments = self._factor_exposures.columns
            B = B.values
            R = R.loc[instruments].values

        if not isinstance(B, ndarray):
            raise TypeError(
                "Only pandas DataFrame / numpy ndarray is supported, but not "
                f"{B.__class__.__name__}"
            )

        cov = B.T @ factor_covariances @ B

        # Add the specific variances into the covariance matrix
        cov[diag_indices_from(cov)] += R

        # Set zero covariance instruments to nan
        valid_instruments = any(cov != 0.0, axis=0)
        cov[~valid_instruments, :] = nan
        cov[:, ~valid_instruments] = nan

        if not self._config.show_all_instruments:
            cov = cov[valid_instruments, :][:, valid_instruments]
            if isinstance(self._factor_exposures, DataFrame):
                instruments = instruments[valid_instruments]

        if isinstance(self._factor_exposures, DataFrame):
            cov = DataFrame(cov, index=instruments, columns=instruments)

        return cov

    def write_directory(self, path: str, format="parquet", **kwargs):
        """
        Write the factor risk model to directory.

        Parameters
        ----------
        path: str
          Destination path.

        format: str
            Supported formats. Default is "parquet". Options
            are "csv", "parquet" and "hdf".

        **kwargs: dict
            Optional keyword arguments for the write operation.
        """
        method_name = f"to_{format}"
        getattr(self._factor_exposures, method_name)(
            join(path, f"factor_exposures.{format}"), **kwargs
        )
        getattr(self._factor_returns, method_name)(
            join(path, f"factor_returns.{format}"), **kwargs
        )
        getattr(self._residual_returns, method_name)(
            join(path, f"residual_returns.{format}"), **kwargs
        )
        with open(join(path, "metadata.json"), mode="w+") as fp:
            json.dump(self.asdict(), fp)

    @classmethod
    def read_directory(cls, path: str, format: str = "parquet", **kwargs):
        """
        Read model from specified directory.

        Parameters
        ----------
        path: str
            Directory to read model from.

        format: str
            Supported formats. Default is "parquet". Options
            are "csv", "parquet" and "hdf".

        **kwargs: dict
            Optional keyword arguments for the write operation.
        """
        method_name = f"read_{format}"
        import pandas

        method = getattr(pandas, method_name)
        factor_exposures = method(join(path, f"factor_exposures.{format}"), **kwargs)
        factor_exposures.index.name = None
        factor_returns = method(join(path, f"factor_returns.{format}"), **kwargs)
        factor_returns.index.name = None
        residual_returns = method(join(path, f"residual_returns.{format}"), **kwargs)
        residual_returns.index.name = None
        with open(join(path, "metadata.json")) as fp:
            metadata = json.load(fp)

        return cls(
            factor_exposures=factor_exposures,
            factor_returns=factor_returns,
            residual_returns=residual_returns,
            **metadata,
        )



from typing import Optional, Union

from numpy import ndarray
from pandas import DataFrame, Series
from sklearn.decomposition import PCA as sklearn_PCA

from ..engine import NumpyEngine
from ..factor_risk_model import FactorRiskModel
from ..regressor import WLS

np = NumpyEngine()


class PCAConfig(FactorRiskModel.ConfigClass):
    """
    PCA statistics model configuration class.


    Parameters
    ----------
    n_components : int
        Number of components.
    demean : Optional[bool]
        Indicate whether to demean before fitting. Default is True.
    speedup: Optional[bool]
        Indicate whether to speed up the computation as much as possible.
        Default is True.
    """

    n_components: Union[int, float, str]
    demean: Optional[bool] = True
    speedup: Optional[bool] = True


class PCA(FactorRiskModel):
    """
    PCA statistics model.
    """

    ConfigClass = PCAConfig

    def __init__(
        self,
        n_components: int,
        demean: Optional[bool] = True,
        speedup: Optional[bool] = True,
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        n_components : int
          Number of components.
        demean : Optional[bool]
          Indicate whether to demean before fitting. Default is True.
        speedup: Optional[bool]
          Indicate whether to speed up the computation as much as possible.
          Default is True.
        """
        super().__init__(
            n_components=n_components, demean=demean, speedup=speedup, **kwargs
        )
        self._model = sklearn_PCA(n_components=n_components)

    def fit(
        self,
        X: Union[ndarray, DataFrame],
        weights: Optional[Union[ndarray, Series]] = None,
    ) -> object:
        """
        Fit the returns into the risk model.

        Parameters
        ----------
        X: pandas.DataFrame or numpy.ndarray
          Instrument returns where the columns are the instruments
          and the index is the date / time in ascending order.
          For example, if there are N instruments and T days of
          returns, the input is with the dimension of (T, N).

        Returns
        -------
        object
          The object itself.
        """
        # First convert all the numpy ndarray type first
        X_fit = self._to_numpy(X)
        weights_fit = self._to_numpy(weights)

        # Normalize the instrument return by full history mean
        if self._config.demean:
            X_mean = np.array(X.mean(axis=0))[np.newaxis, :]
            X_fit = np.subtract(X_fit, X_mean)

        # Remove the instruments without any returns always
        if self._config.speedup:
            # Select the instruments of which the returns are not always 0
            X_reindex = ~np.all(np.abs(X_fit) < 1e-20, axis=0)
            X_fit = X_fit[:, X_reindex]
            if weights_fit is not None:
                weights_fit = weights_fit[X_reindex]

        # Fit with skilearn PCA on the return matrix (T, N)
        self._model.fit(X_fit)
        # N is the number of instruments and T is the number of time frames
        T = X.shape[0]
        N = X.shape[1]
        # Dimension (n, N) where n is the number of instruments
        # Eigenvectors
        U_m = np.array(self._model.components_)
        # Exposure matrix (n, N)
        B = np.multiply(
            U_m,
            np.array(self._model.singular_values_ * (T**0.5))[:, np.newaxis],
        )
        # Factor matrix (T, n)
        wls = WLS()
        wls_result = wls.fit(X=B.T, y=X_fit.T, weights=weights_fit)
        F = wls_result.beta.T
        residual_returns = wls_result.alpha.T

        # Fill back the instruments which don't have any returns
        # with 0.0 exposures and residual returns
        if self._config.speedup:
            B_reindex = np.zeros((B.shape[0], N))
            residual_returns_reindex = np.zeros(X.shape)
            B_reindex[:, X_reindex] = B[:, :]
            residual_returns_reindex[:, X_reindex] = residual_returns[:, :]
            B = B_reindex
            residual_returns = residual_returns_reindex

        # Convert back to dataframe if necessary
        if isinstance(X, DataFrame):
            factor_index = [f"factor_{index + 1}" for index in range(B.shape[0])]
            B = DataFrame(B, index=factor_index, columns=X.columns)
            F = DataFrame(F, index=X.index, columns=factor_index)
            residual_returns = DataFrame(
                residual_returns, index=X.index, columns=X.columns
            )

        # Return itself out
        self._factor_exposures = B
        self._factor_returns = F
        self._residual_returns = residual_returns
        return self



# flake8: noqa
from .bias import compute_bias_statistics, compute_standardized_returns
from .value_at_risk import (
    compute_value_at_risk_breach_statistics,
    compute_value_at_risk_rolling_breach_statistics,
    compute_value_at_risk_threshold,
)
